from typing import Optional, Dict, Any, List, Union

from phidata.asset.file import File
from phidata.asset.table.sql import SqlTable
from phidata.workflow.output_type import OutputType
from phidata.workflow.py import PythonWorkflow, PythonWorkflowArgs, EngineType
from phidata.utils.cli_console import print_info, print_error
from phidata.utils.log import logger


class GetTickersArgs(PythonWorkflowArgs):
    # Args provided by user
    # ['Stock', 'ETF', 'Mutual Fund']
    asset_types: List[str] = []
    use_session: bool = True
    api_key: Optional[str] = None
    # Required
    output_to: Optional[OutputType] = None
    file: Optional[File] = None
    sql_table: Optional[SqlTable] = None
    # Used by this workflow
    tiingo_config: Optional[Dict[str, Any]]


class GetTickers(PythonWorkflow):
    def __init__(
        self,
        asset_types: Optional[List[str]] = None,
        use_session: bool = True,
        api_key: Optional[str] = None,
        output_to: Optional[OutputType] = None,
        file: Optional[File] = None,
        sql_table: Optional[SqlTable] = None,
        engine: Optional[EngineType] = EngineType.PANDAS,
        name: Optional[str] = "get_tickers",
        version: Optional[str] = None,
        enabled: bool = True,
    ):

        super().__init__()
        try:
            self.args: GetTickersArgs = GetTickersArgs(
                asset_types=asset_types or [],
                use_session=use_session,
                api_key=api_key,
                output_to=output_to,
                file=file,
                sql_table=sql_table,
                engine=engine,
                entrypoint=get_tickers_entrypoint,
                name=name or "get_tickers",
                version=version,
                enabled=enabled,
            )
        except Exception as e:
            logger.error("Exception building GetTickers")
            raise


def get_tickers_entrypoint(**kwargs) -> bool:

    args: GetTickersArgs = GetTickersArgs(**kwargs)
    # logger.debug("GetTickersArgs: {}".format(args))

    if args.api_key is None:
        print_error("ApiKey required")
        return False

    if args.engine != EngineType.PANDAS:
        print_error("Only PANDAS engine supported")
        return False

    if args.output_to is None or not isinstance(args.output_to, OutputType):
        print_error("OutputType invalid")
        return False
    output_to = args.output_to

    import pandas as pd
    from tiingo import TiingoClient

    tiingo_config: Dict[str, Any] = {
        "session": args.use_session or True,
        "api_key": args.api_key,
    }
    tiingo_client: TiingoClient = TiingoClient(tiingo_config)

    # tickers_df = tiingo_client.list_tickers(assetTypes=args.asset_types)
    list_tickers_response = tiingo_client.list_stock_tickers()
    # logger.debug("list_tickers_response:")
    # logger.debug("list_tickers_response type: {}".format(type(list_tickers_response)))
    # logger.debug(list_tickers_response[:5])
    tickers_df = pd.DataFrame(list_tickers_response)
    tickers_df.rename(
        columns={
            "assetType": "asset_type",
            "priceCurrency": "price_currency",
            "startDate": "start_date",
            "endDate": "end_date",
        },
        inplace=True,
    )
    # logger.debug("tickers_df:")
    # logger.debug(tickers_df[:5])
    logger.debug("# tickers: {}".format(len(tickers_df)))

    if output_to == OutputType.FILE:
        if args.file is None:
            print_error("File invalid")
            return False

        args.file.write_pandas_df(tickers_df)
    elif output_to == OutputType.SQL_TABLE:
        if args.sql_table is None:
            print_error("SqlTable invalid")
            return False

        args.sql_table.write_pandas_df(tickers_df)

    return True



from collections import OrderedDict
from pathlib import Path
from typing import Optional, Dict, Union, List

from phidata.app import PhidataApp, PhidataAppArgs
from phidata.infra.docker.resource.network import DockerNetwork
from phidata.infra.docker.resource.container import DockerContainer
from phidata.infra.docker.resource.group import (
    DockerResourceGroup,
    DockerBuildContext,
)
from phidata.infra.k8s.create.core.v1.config_map import CreateConfigMap
from phidata.infra.k8s.create.core.v1.container import CreateContainer
from phidata.infra.k8s.create.apps.v1.deployment import CreateDeployment
from phidata.infra.k8s.create.core.v1.secret import CreateSecret
from phidata.infra.k8s.create.core.v1.service import CreateService, ServiceType
from phidata.infra.k8s.create.core.v1.persistent_volume import CreatePersistentVolume
from phidata.infra.k8s.create.core.v1.volume import (
    CreateVolume,
    EmptyDirVolumeSource,
    HostPathVolumeSource,
    VolumeType,
)
from phidata.infra.k8s.create.common.port import CreatePort
from phidata.infra.k8s.create.group import CreateK8sResourceGroup
from phidata.infra.k8s.resource.group import (
    K8sResourceGroup,
    K8sBuildContext,
)
from phidata.utils.common import (
    get_image_str,
    get_default_container_name,
    get_default_configmap_name,
    get_default_secret_name,
    get_default_service_name,
    get_default_volume_name,
    get_default_deploy_name,
    get_default_pod_name,
)
from phidata.utils.enums import ExtendedEnum
from phidata.utils.cli_console import print_error, print_info, print_warning
from phidata.utils.log import logger


class PostgresVolumeType(ExtendedEnum):
    EMPTY_DIR = "EMPTY_DIR"
    HOST_PATH = "HOST_PATH"
    PERSISTENT_VOLUME = "PERSISTENT_VOLUME"
    AWS_EBS = "AWS_EBS"
    GCE_PERSISTENT_DISK = "GCE_PERSISTENT_DISK"


class PostgresDbArgs(PhidataAppArgs):
    name: str = "postgres"
    version: str = "1"
    enabled: bool = True

    # Image Args
    image_name: str = "postgres"
    image_tag: str = "14"
    # Add env variables to container env
    # keys: var name
    # values: var values
    env: Optional[Dict[str, str]] = None

    ## Configure the Postgres env
    postgres_db: Optional[str] = None
    postgres_user: Optional[str] = None
    postgres_password: Optional[str] = None
    pgdata: Optional[str] = "/var/lib/postgresql/data/pgdata"
    postgres_initdb_args: Optional[str] = None
    postgres_initdb_waldir: Optional[str] = None
    postgres_host_auth_method: Optional[str] = None
    postgres_password_file: Optional[str] = None
    postgres_user_file: Optional[str] = None
    postgres_db_file: Optional[str] = None
    postgres_initdb_args_file: Optional[str] = None

    ## Configure the Postgres volume
    create_volume: bool = True
    volume_type: PostgresVolumeType = PostgresVolumeType.EMPTY_DIR
    volume_name: Optional[str] = None
    # Container path to mount the postgres volume
    volume_container_path: str = "/var/lib/postgresql/data"
    # Host path to mount the postgres volume
    # If volume_type = PostgresVolumeType.HOST_PATH
    volume_host_path: Optional[Path] = None

    ## Configure the Postgres container
    container_name: Optional[str] = None
    container_port: int = 5432
    # Only used by the K8sContainer
    container_port_name: str = "pg"
    # Only used by the DockerContainer
    container_host_port: int = 5432
    container_detach: bool = True
    container_auto_remove: bool = True
    container_remove: bool = True

    ## Configure the Postgres deploy
    deploy_name: Optional[str] = None
    pod_name: Optional[str] = None
    # Options used by the K8s ConfigMap used
    # to set the container env variables which are not Secret
    config_map_name: Optional[str] = None
    # Options used by the K8s Secret used
    # to set the container env variables which are Secret
    secret_name: Optional[str] = None
    # Read secrets from a file in yaml format
    secrets_file: Optional[str] = None

    ## Configure the Postgres service
    service_name: Optional[str] = None
    # Service type
    service_type: Optional[ServiceType] = None
    # The port that will be exposed by the service.
    service_port: int = 5432
    # The node_port that will be exposed by the service.
    # If service_type = ServiceType.CLUSTER_IP
    node_port: Optional[int] = None
    # The target_port is the port to access on the pods targeted by the service.
    # It can be the port number or port name on the pod.
    target_port: Optional[Union[str, int]] = None


class PostgresDb(PhidataApp):
    def __init__(
        self,
        name: str = "postgres",
        version: str = "1",
        enabled: bool = True,
        # Image Args
        image_name: str = "postgres",
        image_tag: str = "14",
        # Add env variables to container env
        env: Optional[Dict[str, str]] = None,
        ## Configure the Postgres env,
        postgres_db: Optional[str] = None,
        postgres_user: Optional[str] = None,
        postgres_password: Optional[str] = None,
        pgdata: Optional[str] = "/var/lib/postgresql/data/pgdata",
        postgres_initdb_args: Optional[str] = None,
        postgres_initdb_waldir: Optional[str] = None,
        postgres_host_auth_method: Optional[str] = None,
        postgres_password_file: Optional[str] = None,
        postgres_user_file: Optional[str] = None,
        postgres_db_file: Optional[str] = None,
        postgres_initdb_args_file: Optional[str] = None,
        ## Configure the Postgres volume
        create_volume: bool = True,
        volume_type: PostgresVolumeType = PostgresVolumeType.EMPTY_DIR,
        volume_name: Optional[str] = None,
        # Container path to mount the postgres volume,
        volume_container_path: str = "/var/lib/postgresql/data",
        # Host path to mount the postgres volume,
        # If volume_type = PostgresVolumeType.HOST_PATH,
        volume_host_path: Optional[Path] = None,
        ## Configure the Postgres container,
        container_name: Optional[str] = None,
        container_port: int = 5432,
        # Options used by the Docker Container
        container_host_port: int = 5432,
        container_detach: bool = True,
        container_auto_remove: bool = True,
        container_remove: bool = True,
        # Options used by the K8s Container
        container_port_name: str = "pg",
        # Options used by the K8s Deploy
        deploy_name: Optional[str] = None,
        pod_name: Optional[str] = None,
        # Options used by the K8s ConfigMap used
        # to set the container env variables which are not Secret
        config_map_name: Optional[str] = None,
        # Options used by the K8s Secret used
        # to set the container env variables which are Secret
        secret_name: Optional[str] = None,
        # Read secrets from a file in yaml format
        secrets_file: Optional[str] = None,
        # Options used by the K8s Service
        service_name: Optional[str] = None,
        # Service type
        service_type: Optional[ServiceType] = None,
        # The port that will be exposed by the service.
        service_port: int = 5432,
        # The node_port that will be exposed by the service.
        node_port: Optional[int] = None,
        target_port: Optional[Union[str, int]] = None,
    ):
        super().__init__()
        try:
            self.args: PostgresDbArgs = PostgresDbArgs(
                name=name,
                version=version,
                enabled=enabled,
                image_name=image_name,
                image_tag=image_tag,
                env=env,
                postgres_db=postgres_db,
                postgres_user=postgres_user,
                postgres_password=postgres_password,
                pgdata=pgdata,
                postgres_initdb_args=postgres_initdb_args,
                postgres_initdb_waldir=postgres_initdb_waldir,
                postgres_host_auth_method=postgres_host_auth_method,
                postgres_password_file=postgres_password_file,
                postgres_user_file=postgres_user_file,
                postgres_db_file=postgres_db_file,
                postgres_initdb_args_file=postgres_initdb_args_file,
                create_volume=create_volume,
                volume_type=volume_type,
                volume_name=volume_name,
                volume_container_path=volume_container_path,
                volume_host_path=volume_host_path,
                container_name=container_name,
                container_port=container_port,
                container_port_name=container_port_name,
                container_host_port=container_host_port,
                container_detach=container_detach,
                container_auto_remove=container_auto_remove,
                container_remove=container_remove,
                deploy_name=deploy_name,
                pod_name=pod_name,
                config_map_name=config_map_name,
                secret_name=secret_name,
                secrets_file=secrets_file,
                service_name=service_name,
                service_type=service_type,
                service_port=service_port,
                node_port=node_port,
                target_port=target_port,
            )
        except Exception:
            logger.error(f"Args for {self.__class__.__name__} are not valid")
            raise

    def get_container_name(self) -> str:
        return self.args.container_name or get_default_container_name(self.args.name)

    def get_service_name(self) -> str:
        return self.args.service_name or get_default_service_name(self.args.name)

    def get_connection_url_docker(self):
        user = self.args.postgres_user
        password = self.args.postgres_password
        host = self.get_container_name()
        port = self.args.container_port
        db = self.args.postgres_db
        return f"postgresql://{user}:{password}@{host}:{port}/{db}"

    def get_connection_url_local(self):
        user = self.args.postgres_user
        password = self.args.postgres_password
        host = "localhost"
        port = self.args.container_host_port
        db = self.args.postgres_db
        return f"postgresql://{user}:{password}@{host}:{port}/{db}"

    ######################################################
    ## Docker Resources
    ######################################################

    def get_postgres_docker_rg(
        self, docker_build_context: DockerBuildContext
    ) -> Optional[DockerResourceGroup]:
        logger.debug(f"Init Postgres DockerResourceGroup")

        app_name = self.args.name
        ## Create a dict to set the container env
        container_env = {
            "POSTGRES_USER": self.args.postgres_user,
            "POSTGRES_PASSWORD": self.args.postgres_password,
            "POSTGRES_DB": self.args.postgres_db,
            "PGDATA": self.args.pgdata,
        }
        # Update the container env with user provided env
        if self.args.env is not None and isinstance(self.args.env, dict):
            container_env.update(self.args.env)

        ## Create the container volumes
        container_volumes = {}
        if self.args.create_volume:
            volume_name = self.args.volume_name or get_default_volume_name(app_name)
            if self.args.volume_type == PostgresVolumeType.EMPTY_DIR:
                container_volumes[volume_name] = {
                    "bind": self.args.volume_container_path,
                    "mode": "rw",
                }
            elif self.args.volume_type == PostgresVolumeType.HOST_PATH:
                if self.args.volume_host_path is not None:
                    volume_host_path_str = str(self.args.volume_host_path)
                    container_volumes[volume_host_path_str] = {
                        "bind": self.args.volume_container_path,
                        "mode": "rw",
                    }
                else:
                    print_error("PostgresDb: volume_host_path not provided")
                    return None
            else:
                print_error(f"{self.args.volume_type.value} not supported")
                return None

        ## Create the container
        docker_container = DockerContainer(
            name=self.get_container_name(),
            image=get_image_str(self.args.image_name, self.args.image_tag),
            detach=self.args.container_detach,
            auto_remove=self.args.container_auto_remove,
            remove=self.args.container_remove,
            environment=container_env,
            network=docker_build_context.network,
            ports={
                str(self.args.container_port): self.args.container_host_port,
            },
            volumes=container_volumes,
        )

        docker_rg = DockerResourceGroup(
            name=app_name,
            enabled=self.args.enabled,
            network=DockerNetwork(name=docker_build_context.network),
            containers=[docker_container],
        )
        return docker_rg

    def init_docker_resource_groups(
        self, docker_build_context: DockerBuildContext
    ) -> None:
        docker_rg: Optional[DockerResourceGroup] = self.get_postgres_docker_rg(
            docker_build_context
        )
        # logger.debug("docker_rg:\n{}".format(docker_rg.json(indent=2)))
        if docker_rg is not None:
            if self.docker_resource_groups is None:
                self.docker_resource_groups = OrderedDict()
            self.docker_resource_groups[docker_rg.name] = docker_rg

    ######################################################
    ## K8s Resources
    ######################################################

    def get_postgres_k8s_rg(
        self, k8s_build_context: K8sBuildContext
    ) -> Optional[K8sResourceGroup]:
        logger.debug(f"Init Postgres K8sResourceGroup")

        app_name = self.args.name
        ## Create a ConfigMap to set the container env variables which are not Secret
        container_env_cm = CreateConfigMap(
            cm_name=self.args.config_map_name or get_default_configmap_name(app_name),
            app_name=app_name,
            data={
                "POSTGRES_USER": self.args.postgres_user,
                "POSTGRES_DB": self.args.postgres_db,
                "PGDATA": self.args.pgdata,
                "POSTGRES_INITDB_ARGS": self.args.postgres_initdb_args,
                "POSTGRES_INITDB_WALDIR": self.args.postgres_initdb_waldir,
                "POSTGRES_HOST_AUTH_METHOD": self.args.postgres_host_auth_method,
                "POSTGRES_PASSWORD_FILE": self.args.postgres_password_file,
                "POSTGRES_USER_FILE": self.args.postgres_user_file,
                "POSTGRES_DB_FILE": self.args.postgres_db_file,
                "POSTGRES_INITDB_ARGS_FILE": self.args.postgres_initdb_args_file,
            },
        )
        # Update the container env with user provided env
        if self.args.env is not None and isinstance(self.args.env, dict):
            if container_env_cm.data is not None and isinstance(
                container_env_cm.data, dict
            ):
                container_env_cm.data.update(self.args.env)

        ## Create a Secret to set the container env variables which are Secret
        container_secrets: List[CreateSecret] = []
        if self.args.postgres_password is not None:
            container_secrets.append(
                CreateSecret(
                    secret_name=self.args.secret_name
                    or get_default_secret_name(f"{app_name}-pg"),
                    app_name=app_name,
                    data={
                        "POSTGRES_PASSWORD": self.args.postgres_password,
                    },
                )
            )
        if self.args.secrets_file is not None:
            try:
                import yaml

                secrets_file_path = Path(self.args.secrets_file)
                if secrets_file_path.exists() and secrets_file_path.is_file():
                    secrets_from_file = yaml.safe_load(secrets_file_path.read_text())
                    if secrets_from_file is not None and isinstance(
                        secrets_from_file, dict
                    ):
                        for secret_name, secret_data in secrets_from_file.items():
                            if (
                                secret_name is None
                                or secret_data is None
                                or not isinstance(secret_data, dict)
                            ):
                                print_error("Invalid secret")
                                continue

                            container_secrets.append(
                                CreateSecret(
                                    secret_name=secret_name,
                                    app_name=app_name,
                                    data=secret_data,
                                )
                            )
                    else:
                        print_error("Cannot read secrets_file")
                else:
                    print_error("Invalid secrets_file")
            except Exception as e:
                print_error("Encountered error when reading secrets_file, please fix")
                print_error(e)

        ## Create the container volumes
        container_volumes = []
        if self.args.create_volume:
            volume_name = self.args.volume_name or get_default_volume_name(app_name)
            if self.args.volume_type == PostgresVolumeType.EMPTY_DIR:
                _volume = CreateVolume(
                    volume_name=volume_name,
                    app_name=app_name,
                    mount_path=self.args.volume_container_path,
                    volume_type=VolumeType.EMPTY_DIR,
                )
                container_volumes.append(_volume)
            elif self.args.volume_type == PostgresVolumeType.HOST_PATH:
                if self.args.volume_host_path is not None:
                    volume_host_path_str = str(self.args.volume_host_path)
                    pg_volume = CreateVolume(
                        volume_name=volume_name,
                        app_name=app_name,
                        mount_path=self.args.volume_container_path,
                        volume_type=VolumeType.HOST_PATH,
                        host_path=HostPathVolumeSource(
                            path=volume_host_path_str,
                        ),
                    )
                    container_volumes.append(pg_volume)
                else:
                    print_error("PostgresDb: volume_host_path not provided")
                    return None
            elif self.args.volume_type == PostgresVolumeType.AWS_EBS:
                print_info(f"TO_IMPLEMENT volume_type: {self.args.volume_type.value}")
                return None
            else:
                print_error(f"{self.args.volume_type.value} not supported")
                return None

        ## Create the ports to open
        container_port = CreatePort(
            name=self.args.container_port_name,
            container_port=self.args.container_port,
            service_port=self.args.service_port,
            target_port=self.args.target_port or self.args.container_port_name,
        )

        # If ServiceType == NODE_PORT then validate self.args.node_port is available
        if self.args.service_type == ServiceType.NODE_PORT:
            if (
                self.args.node_port is None
                or self.args.node_port < 30000
                or self.args.node_port > 32767
            ):
                print_error(f"NodePort: {self.args.node_port} invalid")
                print_error(f"Skipping this service")
                return None
            else:
                container_port.node_port = self.args.node_port
        # If ServiceType == LOAD_BALANCER then validate self.args.node_port only IF available
        elif self.args.service_type == ServiceType.LOAD_BALANCER:
            if self.args.node_port is not None:
                if self.args.node_port < 30000 or self.args.node_port > 32767:
                    print_error(f"NodePort: {self.args.node_port} invalid")
                    print_error(f"Skipping this service")
                    return None
                else:
                    container_port.node_port = self.args.node_port
        # else validate self.args.node_port is NOT available
        elif self.args.node_port is not None:
            print_warning(
                f"NodePort: {self.args.node_port} provided without specifying ServiceType as NODE_PORT or LOAD_BALANCER"
            )
            print_warning("NodePort value will be ignored")
            self.args.node_port = None

        ## Create the container
        k8s_container = CreateContainer(
            container_name=self.get_container_name(),
            app_name=app_name,
            image_name=self.args.image_name,
            image_tag=self.args.image_tag,
            envs_from_configmap=[container_env_cm.cm_name],
            envs_from_secret=[secret.secret_name for secret in container_secrets],
            ports=[container_port],
            volumes=container_volumes,
        )

        ## Create the deployment
        k8s_deployment = CreateDeployment(
            deploy_name=self.args.deploy_name or get_default_deploy_name(app_name),
            pod_name=self.args.pod_name or get_default_pod_name(app_name),
            app_name=app_name,
            namespace=k8s_build_context.namespace,
            service_account_name=k8s_build_context.service_account_name,
            containers=[k8s_container],
            volumes=container_volumes,
            labels=k8s_build_context.labels,
        )

        ## Create the service
        k8s_service = CreateService(
            service_name=self.get_service_name(),
            app_name=app_name,
            namespace=k8s_build_context.namespace,
            service_account_name=k8s_build_context.service_account_name,
            service_type=self.args.service_type,
            deployment=k8s_deployment,
            ports=[container_port],
            labels=k8s_build_context.labels,
        )

        ## Create the K8sResourceGroup
        k8s_resource_group = CreateK8sResourceGroup(
            name=app_name,
            enabled=self.args.enabled,
            config_maps=[container_env_cm],
            secrets=container_secrets,
            services=[k8s_service],
            deployments=[k8s_deployment],
        )
        return k8s_resource_group.create()

    def init_k8s_resource_groups(self, k8s_build_context: K8sBuildContext) -> None:
        k8s_rg: Optional[K8sResourceGroup] = self.get_postgres_k8s_rg(k8s_build_context)
        # logger.debug("k8s_rg:\n{}".format(k8s_rg.json(indent=2)))
        if k8s_rg is not None:
            if self.k8s_resource_groups is None:
                self.k8s_resource_groups = OrderedDict()
            self.k8s_resource_groups[k8s_rg.name] = k8s_rg



from typing import List, Optional

from kubernetes.client import CoreV1Api
from kubernetes.client.models.v1_pod import V1Pod
from kubernetes.client.models.v1_pod_list import V1PodList

from phidata.infra.k8s.api_client import K8sApiClient
from phidata.infra.k8s.resource.base import K8sResource


class Pod(K8sResource):
    """
    There are no attributes in the Pod model because we don't create Pods manually.
    This class exists only to read from the K8s cluster.

    References:
        * Doc: https://kubernetes.io/docs/reference/generated/kubernetes-api/v1.23/#pod-v1-core
        * Type: https://github.com/kubernetes-client/python/blob/master/kubernetes/client/models/v1_pod.py
    """

    resource_type: str = "Pod"

    @staticmethod
    def get_from_cluster(
        k8s_client: K8sApiClient, namespace: Optional[str] = None, **kwargs: str
    ) -> Optional[List[V1Pod]]:

        core_v1_api: CoreV1Api = k8s_client.core_v1_api
        pod_list: Optional[V1PodList] = None
        if namespace:
            # logger.debug(f"Getting Pods for ns: {namespace}")
            pod_list = core_v1_api.list_namespaced_pod(namespace=namespace)
        else:
            # logger.debug("Getting SA for all namespaces")
            pod_list = core_v1_api.list_pod_for_all_namespaces()

        pods: Optional[List[V1Pod]] = None
        if pod_list:
            pods = pod_list.items
        # logger.debug(f"pods: {pods}")
        # logger.debug(f"pods type: {type(pods)}")
        return pods



from collections import OrderedDict
from typing import Optional, Dict, Any, List, Union

from phidata.app import PhidataApp, PhidataAppArgs
from phidata.infra.docker.resource.network import DockerNetwork
from phidata.infra.docker.resource.container import DockerContainer
from phidata.infra.docker.resource.group import (
    DockerResourceGroup,
    DockerBuildContext,
)
from phidata.infra.k8s.create.apps.v1.deployment import CreateDeployment
from phidata.infra.k8s.create.core.v1.config_map import CreateConfigMap
from phidata.infra.k8s.create.core.v1.container import CreateContainer
from phidata.infra.k8s.create.core.v1.volume import (
    CreateVolume,
    EmptyDirVolumeSource,
    VolumeType,
)
from phidata.infra.k8s.create.common.port import CreatePort
from phidata.infra.k8s.create.group import CreateK8sResourceGroup
from phidata.infra.k8s.resource.group import (
    K8sResourceGroup,
    K8sBuildContext,
)
from phidata.utils.common import (
    get_default_configmap_name,
    get_default_container_name,
    get_default_deploy_name,
    get_default_pod_name,
)
from phidata.utils.cli_console import print_error
from phidata.utils.log import logger

default_databox_name: str = "databox"


class DataboxArgs(PhidataAppArgs):
    name: str = default_databox_name
    version: str = "1"
    enabled: bool = True

    # Image args
    image_name: str = "phidata/databox"
    image_tag: str = "1"
    # Path to entrypoint script in the image
    entrypoint: Optional[Union[str, List]] = None
    command: Optional[Union[str, List]] = None
    # Add env variables to container env
    # key: var name, value: var value
    env: Optional[Dict[str, str]] = None

    ## Configure python requirements volume
    # Install python dependencies using a requirements.txt file
    install_requirements: bool = False
    requirements_volume_name: str = "databox-requirements-volume"
    # requirements.txt file path in container
    requirements_file_container_path: str = "/requirements.txt"
    # TODO: load requirements_file from git instead of empty dir

    ## Configure the Data volume
    # Creates a data directory on the container
    create_data_volume: bool = False
    data_dir_volume_name: str = "databox-data-volume"
    # Container path to mount the data volume
    data_dir_container_path: str = "/usr/local/data"
    # If True, create the data volume using an empty dir
    use_empty_dir_for_data_volume: bool = True
    # TODO: load storage_dir from s3 instead of empty dir

    ## Configure the Projects volume
    # Create a projects directory on the container
    create_projects_volume: bool = False
    projects_volume_name: str = "databox-projects-volume"
    # Container path to mount the projects volume
    projects_dir_container_path: str = "/usr/local/projects"
    # TODO: load products_dir from git instead of empty dir

    ## Configure Aws credentials
    load_aws_credentials: bool = False

    ## Airflow args
    # Configure the directory containing the airflow dags
    # If use_projects_for_airflow_dags is True
    # set the AIRFLOW__CORE__DAGS_FOLDER to the projects_dir_container_path
    use_projects_for_airflow_dags: bool = True
    # If use_projects_for_airflow_dags is False
    # set the AIRFLOW__CORE__DAGS_FOLDER to the airflow_dags_dir_path
    airflow_dags_dir_path: Optional[str] = None

    # Airflow db args
    wait_for_airflow_db: bool = False
    airflow_db_conn_url: Optional[str] = None
    airflow_db_conn_port: Optional[str] = None
    airflow_db_user: Optional[str] = None
    airflow_db_password: Optional[str] = None
    airflow_schema: Optional[str] = None
    # Airflow db connections.
    # Format expected: dict of { conn_id: conn_url }
    # Converted to env vars: AIRFLOW_CONN__conn_id = conn_url
    db_connections: Optional[Dict] = None

    # Other args
    print_env_on_load: bool = True


class Databox(PhidataApp):
    def __init__(
        self,
        name: str = default_databox_name,
        version: str = "1",
        enabled: bool = True,
        # Image args
        image_name: str = "phidata/databox",
        image_tag: str = "1",
        # Path to entrypoint script in the image
        entrypoint: Optional[Union[str, List]] = None,
        command: Optional[Union[str, List]] = None,
        # Add env variables to container env
        # key: var name, value: var value
        env: Optional[Dict[str, str]] = None,
        ## Configure python requirements volume
        # Install python dependencies using a requirements.txt file
        install_requirements: bool = False,
        requirements_volume_name: str = "databox-requirements-volume",
        # requirements.txt file path in container
        requirements_file_container_path: str = "/requirements.txt",
        ## Configure the Data volume
        # Creates a data directory on the container
        create_data_volume: bool = False,
        data_dir_volume_name: str = "databox-data-volume",
        # Container path to mount the data volume
        data_dir_container_path: str = "/usr/local/data",
        # If True, create the data volume using an empty dir
        use_empty_dir_for_data_volume: bool = True,
        ## Configure the Projects volume
        # Create a projects directory on the container
        create_projects_volume: bool = False,
        projects_volume_name: str = "databox-projects-volume",
        # Container path to mount the projects volume
        projects_dir_container_path: str = "/usr/local/projects",
        ## Configure Aws credentials
        load_aws_credentials: bool = False,
        ## Airflow args
        # Configure the directory containing the airflow dags
        # If use_projects_for_airflow_dags is True
        # set the AIRFLOW__CORE__DAGS_FOLDER to the projects_dir_container_path
        use_projects_for_airflow_dags: bool = True,
        # If use_projects_for_airflow_dags is False
        # set the AIRFLOW__CORE__DAGS_FOLDER to the airflow_dags_dir_path
        airflow_dags_dir_path: Optional[str] = None,
        # Airflow db args
        wait_for_airflow_db: bool = False,
        airflow_db_conn_url: Optional[str] = None,
        airflow_db_conn_port: Optional[str] = None,
        airflow_db_user: Optional[str] = None,
        airflow_db_password: Optional[str] = None,
        airflow_schema: Optional[str] = None,
        # Airflow db connections.
        # Format expected: dict of { conn_id: conn_url }
        # Converted to env vars: AIRFLOW_CONN__conn_id = conn_url
        db_connections: Optional[Dict] = None,
        # Other args
        print_env_on_load: bool = True,
        # Additional args
        # If True, skip resource creation if active resources with the same name exist.
        use_cache: bool = True,
        # If True, log extra debug messages
        use_verbose_logs: bool = False,
    ):

        super().__init__()
        try:
            self.args: DataboxArgs = DataboxArgs(
                name=name,
                version=version,
                enabled=enabled,
                image_name=image_name,
                image_tag=image_tag,
                entrypoint=entrypoint,
                command=command,
                env=env,
                install_requirements=install_requirements,
                requirements_volume_name=requirements_volume_name,
                requirements_file_container_path=requirements_file_container_path,
                create_data_volume=create_data_volume,
                data_dir_volume_name=data_dir_volume_name,
                data_dir_container_path=data_dir_container_path,
                use_empty_dir_for_data_volume=use_empty_dir_for_data_volume,
                create_projects_volume=create_projects_volume,
                projects_volume_name=projects_volume_name,
                projects_dir_container_path=projects_dir_container_path,
                load_aws_credentials=load_aws_credentials,
                use_projects_for_airflow_dags=use_projects_for_airflow_dags,
                airflow_dags_dir_path=airflow_dags_dir_path,
                wait_for_airflow_db=wait_for_airflow_db,
                airflow_db_conn_url=airflow_db_conn_url,
                airflow_db_conn_port=airflow_db_conn_port,
                airflow_db_user=airflow_db_user,
                airflow_db_password=airflow_db_password,
                airflow_schema=airflow_schema,
                db_connections=db_connections,
                print_env_on_load=print_env_on_load,
                use_cache=use_cache,
                use_verbose_logs=use_verbose_logs,
            )
        except Exception as e:
            logger.error(f"Args for {self.__class__.__name__} are not valid")
            raise

    ######################################################
    ## Docker Resources
    ######################################################

    def configure_airflow_on_docker_container(self, container: DockerContainer) -> None:
        """
        Initialize airflow on a docker container
        """

        airflow_env: Dict[str, Any] = {
            "WAIT_FOR_AIRFLOW_DB": str(self.args.wait_for_airflow_db),
            "AIRFLOW_DB_CONN_URL": self.args.airflow_db_conn_url,
            "AIRFLOW_DB_CONN_PORT": self.args.airflow_db_conn_port,
            "AIRFLOW_DB_USER": self.args.airflow_db_user,
            "AIRFLOW_DB_PASSWORD": self.args.airflow_db_password,
            "AIRFLOW_SCHEMA": self.args.airflow_schema,
        }
        # Set the AIRFLOW__CORE__DAGS_FOLDER
        if self.args.create_projects_volume and self.args.use_projects_for_airflow_dags:
            airflow_env[
                "AIRFLOW__CORE__DAGS_FOLDER"
            ] = self.args.projects_dir_container_path
        elif self.args.airflow_dags_dir_path is not None:
            airflow_env["AIRFLOW__CORE__DAGS_FOLDER"] = self.args.airflow_dags_dir_path

        # Set the AIRFLOW__CONN_ variables
        if self.args.db_connections is not None:
            for conn_id, conn_url in self.args.db_connections.items():
                try:
                    af_conn_id = str("AIRFLOW_CONN_{}".format(conn_id)).upper()
                    airflow_env[af_conn_id] = conn_url
                except Exception as e:
                    logger.exception(e)
                    continue
        # Update the container.environment to include airflow_env
        container.environment.update(airflow_env)

    def get_databox_docker_rg(
        self, docker_build_context: DockerBuildContext
    ) -> Optional[DockerResourceGroup]:
        logger.debug(f"Init Databox DockerResourceGroup")

        # Set the Databox Env
        databox_container_env: Dict[str, Any] = {
            # Env variables used by data workflows and data assets
            "PHI_DATA_DIR": self.args.data_dir_container_path,
            "PHI_PROJECTS_DIR": self.args.projects_dir_container_path,
            # Configure python packages to install
            "INSTALL_REQUIREMENTS": str(self.args.install_requirements),
            "REQUIREMENTS_FILE_PATH": self.args.requirements_file_container_path,
            # Print these env variables when the container is loaded
            "PRINT_ENV_ON_LOAD": str(self.args.print_env_on_load),
        }
        # Update the databox env with user provided env
        if self.args.env is not None and isinstance(self.args.env, dict):
            databox_container_env.update(self.args.env)

        # Databox Volumes
        # databox_volumes: A dictionary to configure volumes mounted inside the container.
        # The key is either the host path or a volume name,
        # and the value is a dictionary with 2 keys:
        #   bind - The path to mount the volume inside the container
        #   mode - Either rw to mount the volume read/write, or ro to mount it read-only.
        # For example:
        # {
        #   '/home/user1/': {'bind': '/mnt/vol2', 'mode': 'rw'},
        #   '/var/www': {'bind': '/mnt/vol1', 'mode': 'ro'}
        # }
        databox_volumes = {}
        # Create a volume for the requirements file
        # if install_requirements = True
        if self.args.install_requirements:
            # TODO: load requirements_file from git instead of empty dir
            databox_volumes[self.args.requirements_volume_name] = {
                "bind": self.args.requirements_file_container_path,
                "mode": "rw",
            }
        # Create a volume for the data dir
        # if create_data_volume = True
        if self.args.create_data_volume:
            # TODO: load storage_dir from s3 instead of empty dir
            if self.args.use_empty_dir_for_data_volume:
                databox_volumes[self.args.data_dir_volume_name] = {
                    "bind": self.args.data_dir_container_path,
                    "mode": "rw",
                }
        # Create a volume for the projects dir
        # if create_projects_volume = True
        if self.args.create_projects_volume:
            # TODO: load products_dir from git instead of empty dir
            databox_volumes[self.args.projects_volume_name] = {
                "bind": self.args.projects_dir_container_path,
                "mode": "rw",
            }

        # Create the databox container
        databox_docker_container = DockerContainer(
            name=self.args.name,
            image=f"{self.args.image_name}:{self.args.image_tag}",
            command=self.args.command,
            auto_remove=True,
            detach=True,
            entrypoint=self.args.entrypoint,
            environment=databox_container_env,
            remove=True,
            stdin_open=True,
            tty=True,
            network=docker_build_context.network,
            volumes=databox_volumes,
            use_cache=self.args.use_cache,
            use_verbose_logs=self.args.use_verbose_logs,
        )
        # Configure airflow on container
        self.configure_airflow_on_docker_container(databox_docker_container)

        databox_docker_resource_group = DockerResourceGroup(
            name=self.args.name,
            enabled=self.args.enabled,
            network=DockerNetwork(name=docker_build_context.network),
            containers=[databox_docker_container],
        )
        return databox_docker_resource_group

    def init_docker_resource_groups(
        self, docker_build_context: DockerBuildContext
    ) -> None:
        databox_docker_rg: Optional[DockerResourceGroup] = self.get_databox_docker_rg(
            docker_build_context
        )
        if databox_docker_rg is not None:
            if self.docker_resource_groups is None:
                self.docker_resource_groups = OrderedDict()
            self.docker_resource_groups[databox_docker_rg.name] = databox_docker_rg

    ######################################################
    ## K8s Resources
    ######################################################

    def configure_airflow_on_k8s_container(
        self, container: CreateContainer, k8s_resource_group: CreateK8sResourceGroup
    ) -> None:
        """
        Initialize airflow on a k8s container
        """

        airflow_cm = CreateConfigMap(
            cm_name=get_default_configmap_name("databox-airflow"),
            app_name=self.args.name,
            data={
                "WAIT_FOR_AIRFLOW_DB": str(self.args.wait_for_airflow_db),
                "AIRFLOW_DB_CONN_URL": self.args.airflow_db_conn_url,
                "AIRFLOW_DB_CONN_PORT": self.args.airflow_db_conn_port,
                "AIRFLOW_DB_USER": self.args.airflow_db_user,
                "AIRFLOW_DB_PASSWORD": self.args.airflow_db_password,
                "AIRFLOW_SCHEMA": self.args.airflow_schema,
            },
        )

        # Set the AIRFLOW__CORE__DAGS_FOLDER
        if self.args.create_projects_volume and self.args.use_projects_for_airflow_dags:
            airflow_cm.data[
                "AIRFLOW__CORE__DAGS_FOLDER"
            ] = self.args.projects_dir_container_path
        elif self.args.airflow_dags_dir_path is not None:
            airflow_cm.data[
                "AIRFLOW__CORE__DAGS_FOLDER"
            ] = self.args.airflow_dags_dir_path

        # Set the AIRFLOW__CONN_ variables
        if self.args.db_connections is not None:
            for conn_id, conn_url in self.args.db_connections.items():
                try:
                    af_conn_id = str("AIRFLOW_CONN_{}".format(conn_id)).upper()
                    airflow_cm.data[af_conn_id] = conn_url
                except Exception as e:
                    logger.exception(e)
                    continue

        # Add airflow_cm to the container env
        if container.envs_from_configmap is None:
            container.envs_from_configmap = []
        container.envs_from_configmap.append(airflow_cm.cm_name)

        # Add airflow_cm to k8s_resource_group
        if k8s_resource_group.config_maps is None:
            k8s_resource_group.config_maps = []
        k8s_resource_group.config_maps.append(airflow_cm)

    def get_databox_k8s_rg(
        self, k8s_build_context: K8sBuildContext
    ) -> Optional[K8sResourceGroup]:
        logger.debug(f"Init Databox K8sResourceGroup")

        # Set the Databox Env
        # Databox Config-Map is a dict of environment variables which are not secrets.
        # The keys of this configmap will become env variables in the databox container
        databox_cm = CreateConfigMap(
            cm_name=get_default_configmap_name(self.args.name),
            app_name=self.args.name,
            data={
                # Env variables used by data workflows and data assets
                "PHI_DATA_DIR": self.args.data_dir_container_path,
                "PHI_PROJECTS_DIR": self.args.projects_dir_container_path,
                # Configure python packages to install
                "INSTALL_REQUIREMENTS": str(self.args.install_requirements),
                "REQUIREMENTS_FILE_PATH": self.args.requirements_file_container_path,
                # Print these env variables when the container is loaded
                "PRINT_ENV_ON_LOAD": str(self.args.print_env_on_load),
            },
        )
        # Update the databox env with user provided env
        if self.args.env is not None and isinstance(self.args.env, dict):
            databox_cm.data.update(self.args.env)

        # Databox Volumes
        databox_volumes = []
        # Create a volume for the requirements file
        # if install_requirements = True
        if self.args.install_requirements:
            requirements_volume = CreateVolume(
                volume_name=self.args.requirements_volume_name,
                app_name=self.args.name,
                mount_path=self.args.requirements_file_container_path,
                volume_type=VolumeType.EMPTY_DIR,
                empty_dir=EmptyDirVolumeSource(),
            )
            databox_volumes.append(requirements_volume)
        # Create a volume for the data dir
        # if create_data_volume = True
        if self.args.create_data_volume:
            if self.args.use_empty_dir_for_data_volume:
                data_volume = CreateVolume(
                    volume_name=self.args.data_dir_volume_name,
                    app_name=self.args.name,
                    mount_path=self.args.data_dir_container_path,
                    volume_type=VolumeType.EMPTY_DIR,
                    empty_dir=EmptyDirVolumeSource(),
                )
                databox_volumes.append(data_volume)
        # Create a volume for the projects dir
        # if create_projects_volume = True
        if self.args.create_projects_volume:
            projects_volume = CreateVolume(
                volume_name=self.args.projects_volume_name,
                app_name=self.args.name,
                mount_path=self.args.projects_dir_container_path,
                volume_type=VolumeType.EMPTY_DIR,
                empty_dir=EmptyDirVolumeSource(),
            )
            databox_volumes.append(projects_volume)

        # Create the databox container
        databox_k8s_container = CreateContainer(
            container_name=get_default_container_name(self.args.name),
            app_name=self.args.name,
            image_name=self.args.image_name,
            image_tag=self.args.image_tag,
            # Equivalent to docker images CMD
            args=self.args.command,
            # Equivalent to docker images ENTRYPOINT
            command=self.args.entrypoint,
            envs_from_configmap=[databox_cm.cm_name],
            volumes=databox_volumes,
        )
        databox_deployment = CreateDeployment(
            deploy_name=get_default_deploy_name(self.args.name),
            pod_name=get_default_pod_name(self.args.name),
            app_name=self.args.name,
            namespace=k8s_build_context.namespace,
            service_account_name=k8s_build_context.service_account_name,
            containers=[databox_k8s_container],
            volumes=databox_volumes,
            labels=k8s_build_context.labels,
        )
        databox_k8s_resource_group = CreateK8sResourceGroup(
            name=self.args.name,
            enabled=self.args.enabled,
            config_maps=[databox_cm],
            deployments=[databox_deployment],
        )
        # Configure airflow on container
        self.configure_airflow_on_k8s_container(
            databox_k8s_container, databox_k8s_resource_group
        )
        return databox_k8s_resource_group.create()

    def init_k8s_resource_groups(self, k8s_build_context: K8sBuildContext) -> None:
        databox_k8s_rg: Optional[K8sResourceGroup] = self.get_databox_k8s_rg(
            k8s_build_context
        )
        # logger.debug("databox_k8s_rg: {}".format(databox_k8s_rg))
        if databox_k8s_rg is not None:
            if self.k8s_resource_groups is None:
                self.k8s_resource_groups = OrderedDict()
            self.k8s_resource_groups[databox_k8s_rg.name] = databox_k8s_rg



from typing import Optional, Dict, Any, List, Union

from phidata.asset.file import File
from phidata.asset.table.sql import SqlTable
from phidata.workflow.output_type import OutputType
from phidata.workflow.py import PythonWorkflow, PythonWorkflowArgs, EngineType
from phidata.utils.cli_console import print_info, print_error
from phidata.utils.log import logger


class GetTickerPriceArgs(PythonWorkflowArgs):
    # Args provided by user
    tickers: Optional[Union[str, List[str]]] = None
    # start_date: Start of ticker range in YYYY-MM-DD format.
    start_date: Optional[str] = None
    # end_date: End of ticker range in YYYY-MM-DD format.
    end_date: Optional[str] = None
    frequency: str = "daily"
    response_format: str = "json"
    # valid options = {'open', 'high', 'low', 'close', 'volume',
    # 'adjOpen', 'adjHigh', 'adjLow', 'adjClose', 'adjVolume',
    # 'divCash', 'splitFactor'}
    # metric: Optional[str] = None
    # sort: Optional[str]
    # columns: Optional[List[str]]
    use_session: bool = True
    api_key: Optional[str] = None
    # Required
    output_to: Optional[OutputType] = None
    file: Optional[File] = None
    sql_table: Optional[SqlTable] = None
    # Used by this workflow
    tiingo_config: Optional[Dict[str, Any]]


class GetTickerPrice(PythonWorkflow):
    def __init__(
        self,
        tickers: Optional[Union[str, List[str]]] = None,
        start_date: Optional[str] = None,
        end_date: Optional[str] = None,
        frequency: str = "daily",
        response_format: str = "json",
        use_session: bool = True,
        api_key: Optional[str] = None,
        output_to: Optional[OutputType] = None,
        file: Optional[File] = None,
        sql_table: Optional[SqlTable] = None,
        engine: Optional[EngineType] = EngineType.PANDAS,
        name: Optional[str] = "get_ticker_price",
        version: Optional[str] = None,
        enabled: bool = True,
    ):

        super().__init__()
        try:
            self.args: GetTickerPriceArgs = GetTickerPriceArgs(
                tickers=tickers,
                start_date=start_date,
                end_date=end_date,
                frequency=frequency,
                response_format=response_format,
                use_session=use_session,
                api_key=api_key,
                output_to=output_to,
                file=file,
                sql_table=sql_table,
                engine=engine,
                entrypoint=get_ticker_price_entrypoint,
                name=name or "get_ticker_price",
                version=version,
                enabled=enabled,
            )
        except Exception as e:
            logger.error("Exception building GetTickerPrice")
            raise


def get_ticker_price_entrypoint(**kwargs) -> bool:

    args: GetTickerPriceArgs = GetTickerPriceArgs(**kwargs)
    # logger.debug("GetTickerPriceArgs: {}".format(args))

    if args.api_key is None:
        print_error("ApiKey required")
        return False

    if args.tickers is None:
        print_error("Ticker required")
        return False

    if args.response_format != "json":
        print_error("Only JSON response supported")
        return False

    if args.engine != EngineType.PANDAS:
        print_error("Only PANDAS engine supported")
        return False

    if args.output_to is None or not isinstance(args.output_to, OutputType):
        print_error("OutputType invalid")
        return False
    output_to = args.output_to

    import pandas as pd
    from tiingo import TiingoClient

    tiingo_config: Dict[str, Any] = {
        "session": args.use_session or True,
        "api_key": args.api_key,
    }
    tiingo_client: TiingoClient = TiingoClient(tiingo_config)

    tickers = args.tickers
    ticker_price_df = pd.DataFrame()
    if isinstance(tickers, list):
        for ticker in tickers:
            _price_df: pd.DataFrame = tiingo_client.get_dataframe(
                tickers=ticker,
                startDate=args.start_date,
                endDate=args.end_date,
                frequency=args.frequency,
                fmt=args.response_format,
            )
            # logger.debug("_price_df:")
            # logger.debug(_price_df)
            _price_df["ticker"] = ticker
            ticker_price_df = ticker_price_df.append(_price_df)
    else:
        ticker_price_df = tiingo_client.get_dataframe(
            tickers=tickers,
            startDate=args.start_date,
            endDate=args.end_date,
            frequency=args.frequency,
            fmt=args.response_format,
        )
        ticker_price_df["ticker"] = args.tickers

    ticker_price_df.reset_index(inplace=True)
    ticker_price_df.set_index(["date", "ticker"], inplace=True)
    # logger.debug("ticker_price_df:")
    # logger.debug(ticker_price_df)
    # logger.debug("index: {}".format(ticker_price_df.index))

    if output_to == OutputType.FILE:
        if args.file is None:
            print_error("File invalid")
            return False

        args.file.write_pandas_df(ticker_price_df)
    elif output_to == OutputType.SQL_TABLE:
        if args.sql_table is None:
            print_error("SqlTable invalid")
            return False

        args.sql_table.write_pandas_df(ticker_price_df)

    return True



from collections import OrderedDict
from typing import Dict, List, Type, Union

from phidata.infra.aws.resource.base import AwsResource
from phidata.infra.aws.resource.acm.certificate import AcmCertificate
from phidata.infra.aws.resource.cloudformation import CloudFormationStack
from phidata.infra.aws.resource.eks.cluster import EksCluster
from phidata.infra.aws.resource.eks.fargate_profile import EksFargateProfile
from phidata.infra.aws.resource.eks.node_group import EksNodeGroup
from phidata.infra.aws.resource.iam.role import IamRole
from phidata.infra.aws.resource.iam.policy import IamPolicy
from phidata.infra.aws.resource.glue.crawler import GlueCrawlerResource
from phidata.infra.aws.resource.s3 import S3Bucket

# Use this as a type for an object which can hold any Aws Resource
AwsResourceType = Union[
    AcmCertificate,
    CloudFormationStack,
    EksCluster,
    EksFargateProfile,
    EksNodeGroup,
    IamRole,
    IamPolicy,
    GlueCrawlerResource,
    S3Bucket,
]

# Use this as an ordered list to iterate over all Aws Resource Classes
# This list is the order in which resources should be installed as well.
AwsResourceTypeList: List[Type[AwsResource]] = [
    IamRole,
    IamPolicy,
    S3Bucket,
    AcmCertificate,
    GlueCrawlerResource,
    CloudFormationStack,
    EksCluster,
    EksFargateProfile,
    EksNodeGroup,
]

# Map Aws resource alias' to their type
_aws_resource_type_names: Dict[str, Type[AwsResource]] = {
    aws_type.__name__.lower(): aws_type for aws_type in AwsResourceTypeList
}
_aws_resource_type_aliases: Dict[str, Type[AwsResource]] = {
    "s3": S3Bucket,
}

AwsResourceAliasToTypeMap: Dict[str, Type[AwsResource]] = dict(
    **_aws_resource_type_names, **_aws_resource_type_aliases
)

# Maps each AwsResource to an install weight
# lower weight AwsResource(s) get installed first
AwsResourceInstallOrder: Dict[str, int] = OrderedDict(
    {
        resource_type.__name__: idx
        for idx, resource_type in enumerate(AwsResourceTypeList, start=1)
    }
)



from collections import OrderedDict
from pathlib import Path
from typing import Optional, Dict, Any, List, Union

from pydantic import BaseModel

from phidata.app import PhidataApp, PhidataAppArgs
from phidata.infra.docker.resource.network import DockerNetwork
from phidata.infra.docker.resource.container import DockerContainer
from phidata.infra.docker.resource.group import (
    DockerResourceGroup,
    DockerBuildContext,
)
from phidata.infra.k8s.create.apps.v1.deployment import CreateDeployment
from phidata.infra.k8s.create.core.v1.config_map import CreateConfigMap
from phidata.infra.k8s.create.core.v1.container import CreateContainer
from phidata.infra.k8s.create.core.v1.volume import (
    CreateVolume,
    EmptyDirVolumeSource,
    HostPathVolumeSource,
    VolumeType,
)
from phidata.infra.k8s.create.common.port import CreatePort
from phidata.infra.k8s.create.group import CreateK8sResourceGroup
from phidata.infra.k8s.resource.group import (
    K8sResourceGroup,
    K8sBuildContext,
)
from phidata.utils.common import get_default_configmap_name
from phidata.utils.cli_console import print_error, print_info
from phidata.utils.log import logger

default_devbox_name: str = "devbox"


class DevboxDevModeArgs(BaseModel):
    install_phidata_dev: bool = True
    phidata_volume_name: str = "devbox-phidata-volume"
    phidata_dir_path: Path = Path.home().joinpath("p", "philab", "phidata")
    phidata_dir_container_path: str = "/phidata"


class DevboxArgs(PhidataAppArgs):
    name: str = default_devbox_name
    version: str = "1"
    enabled: bool = True

    # Image args
    image_name: str = "phidata/devbox"
    image_tag: str = "1"
    # Path to entrypoint script for the image
    entrypoint: Optional[Union[str, List]] = None
    command: Optional[Union[str, List]] = None
    # Add env variables to container env
    # key: var name, value: var value
    env: Optional[Dict[str, str]] = None

    # Mount the workspace directory on the container
    mount_workspace: bool = True
    workspace_volume_name: str = "devbox-workspace-volume"
    # Path to mount the workspace volume under
    # This is the parent directory for the workspace on the container
    # i.e. the ws is mounted as a subdir in this dir
    # eg: if ws name is: idata, workspace path would be: /usr/local/devbox/idata
    workspace_parent_container_path: str = "/usr/local/devbox"

    # Install python dependencies using a requirements.txt file
    install_requirements: bool = True
    # Path to the requirements.txt file relative to the workspace root
    requirements_file_path: str = "requirements.txt"

    # Mount Aws config on the container
    mount_aws_config: bool = False
    # Aws config dir on the host
    aws_config_path: Path = Path.home().resolve().joinpath(".aws")
    # Aws config dir on the container
    aws_config_container_path: str = "/usr/local/devbox/.aws"

    # Airflow args
    # If init_airflow = True, this devbox initializes airflow
    init_airflow: bool = True
    # If use_products_as_airflow_dags = True
    # set the AIRFLOW__CORE__DAGS_FOLDER to the products_dir
    use_products_as_airflow_dags: bool = True
    # If use_products_as_airflow_dags = False
    # set the AIRFLOW__CORE__DAGS_FOLDER to the airflow_dags_path
    # airflow_dags_path is the directory in the container containing the airflow dags
    airflow_dags_path: Optional[str] = None
    # Creates an airflow user with username: test, pass: test
    create_airflow_test_user: bool = False

    # Airflow db args
    # If True, initialize the airflow_db on this devbox
    # If None, value is derived from init_airflow i.e. initialize the airflow_db if init_airflow = True
    #   Locally, airflow_db uses sqllite
    # If using the devbox with an external Airflow db, set init_airflow_db = False
    init_airflow_db: Optional[bool] = None
    wait_for_airflow_db: bool = False
    airflow_db_conn_url: Optional[str] = "airflow-pg"
    airflow_db_conn_port: Optional[str] = "5432"
    airflow_db_user: Optional[str] = "airflow"
    airflow_db_password: Optional[str] = "airflow"
    airflow_schema: Optional[str] = "airflow"
    # Airflow db connections in the format { conn_id: conn_url }
    # converted to env var: AIRFLOW_CONN__conn_id = conn_url
    db_connections: Optional[Dict] = None

    # Airflow scheduler args
    # Init Airflow scheduler as a deamon process
    init_airflow_scheduler: bool = False

    # Airflow webserver args
    # Init Airflow webserver when the container starts
    init_airflow_webserver: bool = False
    # webserver port on the container
    airflow_webserver_container_port: int = 8080
    # webserver port on the host machine
    airflow_webserver_host_port: int = 8080
    # K8s args
    # webserver port name
    airflow_webserver_port_name: str = "http"

    # Other airflow args
    load_examples: bool = False
    # Create a soft link from airflow_dir to airflow_home
    # Useful when debugging the airflow conf
    link_airflow_home: bool = False
    # Path to the airflow dir relative to the workspace root
    airflow_dir: str = "airflow"
    # Path to airflow home
    airflow_home: str = "/usr/local/airflow"

    # Other args
    print_env_on_load: bool = True
    dev_mode: Optional[DevboxDevModeArgs] = None


class Devbox(PhidataApp):
    def __init__(
        self,
        name: str = default_devbox_name,
        version: str = "1",
        enabled: bool = True,
        # Image args,
        image_name: str = "phidata/devbox",
        image_tag: str = "1",
        # Path to entrypoint script for the image,
        entrypoint: Optional[Union[str, List]] = None,
        command: Optional[Union[str, List]] = None,
        # Add env variables to container env,
        # key: var name, value: var value,
        env: Optional[Dict[str, str]] = None,
        # Mount the workspace directory on the container,
        mount_workspace: bool = True,
        workspace_volume_name: str = "devbox-workspace-volume",
        # Path to mount the workspace volume under,
        # This is the parent directory for the workspace on the container,
        # i.e. the ws is mounted as a subdir in this dir,
        # eg: if ws name is: idata, workspace path would be: /usr/local/devbox/idata,
        workspace_parent_container_path: str = "/usr/local/devbox",
        # Install python dependencies using a requirements.txt file,
        install_requirements: bool = True,
        # Path to the requirements.txt file relative to the workspace root,
        requirements_file_path: str = "requirements.txt",
        # Mount Aws config on the container,
        mount_aws_config: bool = False,
        # Aws config dir on the host,
        aws_config_path: Path = Path.home().resolve().joinpath(".aws"),
        # Aws config dir on the container,
        aws_config_container_path: str = "/usr/local/devbox/.aws",
        # Airflow args,
        # If init_airflow = True, this devbox initializes airflow,
        init_airflow: bool = True,
        # If use_products_as_airflow_dags = True,
        # set the AIRFLOW__CORE__DAGS_FOLDER to the products_dir,
        use_products_as_airflow_dags: bool = True,
        # If use_products_as_airflow_dags = False,
        # set the AIRFLOW__CORE__DAGS_FOLDER to the airflow_dags_path,
        # airflow_dags_path is the directory in the container containing the airflow dags,
        airflow_dags_path: Optional[str] = None,
        # Creates an airflow user with username: test, pass: test,
        create_airflow_test_user: bool = False,
        # Airflow db args,
        # If True, initialize the airflow_db on this devbox,
        # If None, value is derived from init_airflow i.e. initialize the airflow_db if init_airflow = True,
        #   Locally, airflow_db uses sqllite,
        # If using the devbox with an external Airflow db, set init_airflow_db = False,
        init_airflow_db: Optional[bool] = None,
        wait_for_airflow_db: bool = False,
        airflow_db_conn_url: Optional[str] = "airflow-pg",
        airflow_db_conn_port: Optional[str] = "5432",
        airflow_db_user: Optional[str] = "airflow",
        airflow_db_password: Optional[str] = "airflow",
        airflow_schema: Optional[str] = "airflow",
        # Airflow db connections in the format { conn_id: conn_url },
        # converted to env var: AIRFLOW_CONN__conn_id = conn_url,
        db_connections: Optional[Dict] = None,
        # Airflow scheduler args,
        # Init Airflow scheduler as a deamon process,
        init_airflow_scheduler: bool = False,
        # Airflow webserver args,
        # Init Airflow webserver when the container starts,
        init_airflow_webserver: bool = False,
        # webserver port on the container,
        airflow_webserver_container_port: int = 8080,
        # webserver port on the host machine,
        airflow_webserver_host_port: int = 8080,
        # K8s args,
        # webserver port name,
        airflow_webserver_port_name: str = "http",
        # Other airflow args,
        load_examples: bool = False,
        # Create a soft link from airflow_dir to airflow_home,
        # Useful when debugging the airflow conf,
        link_airflow_home: bool = False,
        # Path to the airflow dir relative to the workspace root,
        airflow_dir: str = "airflow",
        # Path to airflow home,
        airflow_home: str = "/usr/local/airflow",
        # Other args,
        print_env_on_load: bool = True,
        dev_mode: Optional[DevboxDevModeArgs] = None,
        # If True, skip resource creation if active resources with the same name exist.
        use_cache: bool = True,
        # If True, log extra debug messages
        use_verbose_logs: bool = False,
    ):

        super().__init__()
        try:
            self.args: DevboxArgs = DevboxArgs(
                name=name,
                version=version,
                enabled=enabled,
                image_name=image_name,
                image_tag=image_tag,
                entrypoint=entrypoint,
                command=command,
                env=env,
                mount_workspace=mount_workspace,
                workspace_volume_name=workspace_volume_name,
                workspace_parent_container_path=workspace_parent_container_path,
                install_requirements=install_requirements,
                requirements_file_path=requirements_file_path,
                mount_aws_config=mount_aws_config,
                aws_config_path=aws_config_path,
                aws_config_container_path=aws_config_container_path,
                init_airflow=init_airflow,
                use_products_as_airflow_dags=use_products_as_airflow_dags,
                airflow_dags_path=airflow_dags_path,
                create_airflow_test_user=create_airflow_test_user,
                init_airflow_db=init_airflow_db,
                wait_for_airflow_db=wait_for_airflow_db,
                airflow_db_conn_url=airflow_db_conn_url,
                airflow_db_conn_port=airflow_db_conn_port,
                airflow_db_user=airflow_db_user,
                airflow_db_password=airflow_db_password,
                airflow_schema=airflow_schema,
                db_connections=db_connections,
                init_airflow_scheduler=init_airflow_scheduler,
                init_airflow_webserver=init_airflow_webserver,
                airflow_webserver_container_port=airflow_webserver_container_port,
                airflow_webserver_host_port=airflow_webserver_host_port,
                airflow_webserver_port_name=airflow_webserver_port_name,
                load_examples=load_examples,
                link_airflow_home=link_airflow_home,
                airflow_dir=airflow_dir,
                airflow_home=airflow_home,
                print_env_on_load=print_env_on_load,
                dev_mode=dev_mode,
                use_cache=use_cache,
                use_verbose_logs=use_verbose_logs,
            )
        except Exception as e:
            logger.error(f"Args for {self.__class__.__name__} are not valid")
            raise

    ######################################################
    ## Docker Resources
    ######################################################

    def init_airflow_on_docker_container(self, container: DockerContainer) -> None:
        """
        Initialize airflow on a docker container
        """

        if not self.args.init_airflow:
            return

        # Update init_airflow_db arg if None
        if self.args.init_airflow_db is None:
            self.args.init_airflow_db = True

        # Workspace paths
        workspace_name = self.workspace_root_path.stem
        workspace_root_container_path = Path(
            self.args.workspace_parent_container_path
        ).joinpath(workspace_name)
        products_dir_container_path = workspace_root_container_path.joinpath(
            self.products_dir
        )
        airflow_dir_container_path = workspace_root_container_path.joinpath(
            self.args.airflow_dir
        )

        airflow_env: Dict[str, Any] = {
            "INIT_AIRFLOW": str(self.args.init_airflow),
            "INIT_AIRFLOW_DB": str(self.args.init_airflow_db),
            "WAIT_FOR_AIRFLOW_DB": str(self.args.wait_for_airflow_db),
            "AIRFLOW_DB_CONN_URL": self.args.airflow_db_conn_url,
            "AIRFLOW_DB_CONN_PORT": self.args.airflow_db_conn_port,
            "AIRFLOW_DB_USER": self.args.airflow_db_user,
            "AIRFLOW_DB_PASSWORD": self.args.airflow_db_password,
            "AIRFLOW_SCHEMA": self.args.airflow_schema,
            "INIT_AIRFLOW_SCHEDULER": str(self.args.init_airflow_scheduler),
            "INIT_AIRFLOW_WEBSERVER": str(self.args.init_airflow_webserver),
            "AIRFLOW__CORE__LOAD_EXAMPLES": str(self.args.load_examples),
            "LINK_AIRFLOW_HOME": str(self.args.link_airflow_home),
            "AIRFLOW_DIR": str(airflow_dir_container_path),
            "AIRFLOW_HOME": self.args.airflow_home,
            "CREATE_AIRFLOW_TEST_USER": str(self.args.create_airflow_test_user),
        }
        # Set the AIRFLOW__CORE__DAGS_FOLDER
        if self.args.mount_workspace and self.args.use_products_as_airflow_dags:
            airflow_env["AIRFLOW__CORE__DAGS_FOLDER"] = str(products_dir_container_path)
        elif self.args.airflow_dags_path is not None:
            airflow_env["AIRFLOW__CORE__DAGS_FOLDER"] = self.args.airflow_dags_path

        # Set the AIRFLOW__CONN_ variables
        if self.args.db_connections is not None:
            for conn_id, conn_url in self.args.db_connections.items():
                try:
                    af_conn_id = str("AIRFLOW_CONN_{}".format(conn_id)).upper()
                    airflow_env[af_conn_id] = conn_url
                except Exception as e:
                    logger.exception(e)
                    continue
        # Update the container.environment to include airflow_env
        container.environment.update(airflow_env)

        # Open the airflow webserver port if init_airflow_webserver = True
        if self.args.init_airflow_webserver:
            ws_port: Dict[str, Any] = {
                str(
                    self.args.airflow_webserver_container_port
                ): self.args.airflow_webserver_host_port,
            }
            if container.ports is None:
                container.ports = {}
            container.ports.update(ws_port)

    def get_devbox_docker_rg(
        self, docker_build_context: DockerBuildContext
    ) -> Optional[DockerResourceGroup]:
        logger.debug(f"Init Devbox DockerResourceGroup")

        # Workspace paths
        workspace_name = self.workspace_root_path.stem
        workspace_root_container_path = Path(
            self.args.workspace_parent_container_path
        ).joinpath(workspace_name)
        requirements_file_container_path = workspace_root_container_path.joinpath(
            self.args.requirements_file_path
        )

        scripts_dir_container_path = workspace_root_container_path.joinpath(
            self.scripts_dir
        )
        storage_dir_container_path = workspace_root_container_path.joinpath(
            self.storage_dir
        )
        meta_dir_container_path = workspace_root_container_path.joinpath(self.meta_dir)
        products_dir_container_path = workspace_root_container_path.joinpath(
            self.products_dir
        )
        notebooks_dir_container_path = workspace_root_container_path.joinpath(
            self.notebooks_dir
        )
        workspace_config_dir_container_path = workspace_root_container_path.joinpath(
            self.workspace_config_dir
        )

        # Set the Devbox Env
        devbox_container_env: Dict[str, Any] = {
            # Env variables used by data workflows and data assets
            "PHI_WORKSPACE_PARENT": str(self.args.workspace_parent_container_path),
            "PHI_WORKSPACE_ROOT": str(workspace_root_container_path),
            "PHI_SCRIPTS_DIR": str(scripts_dir_container_path),
            "PHI_STORAGE_DIR": str(storage_dir_container_path),
            "PHI_META_DIR": str(meta_dir_container_path),
            "PHI_PRODUCTS_DIR": str(products_dir_container_path),
            "PHI_NOTEBOOKS_DIR": str(notebooks_dir_container_path),
            "PHI_WORKSPACE_CONFIG_DIR": str(workspace_config_dir_container_path),
            "INSTALL_REQUIREMENTS": str(self.args.install_requirements),
            "REQUIREMENTS_FILE_PATH": str(requirements_file_container_path),
            # Print these env variables when the container is loaded
            "PRINT_ENV_ON_LOAD": str(self.args.print_env_on_load),
            # Dev mode args
            "INSTALL_PHIDATA_DEV": str(self.args.dev_mode.install_phidata_dev)
            if self.args.dev_mode is not None
            else str(False),
            "PHIDATA_DIR_PATH": self.args.dev_mode.phidata_dir_container_path
            if self.args.dev_mode is not None
            else str(False),
        }
        # Update the devbox env with user provided env
        if self.args.env is not None and isinstance(self.args.env, dict):
            devbox_container_env.update(self.args.env)

        # Devbox Volumes
        # devbox_volumes: A dictionary to configure volumes mounted inside the container.
        # The key is either the host path or a volume name,
        # and the value is a dictionary with 2 keys:
        #   bind - The path to mount the volume inside the container
        #   mode - Either rw to mount the volume read/write, or ro to mount it read-only.
        # For example:
        # {
        #   '/home/user1/': {'bind': '/mnt/vol2', 'mode': 'rw'},
        #   '/var/www': {'bind': '/mnt/vol1', 'mode': 'ro'}
        # }
        devbox_volumes = {}
        # Create a volume for the workspace dir
        if self.args.mount_workspace:
            workspace_root_path_str = str(self.workspace_root_path)
            workspace_root_container_path_str = str(workspace_root_container_path)
            logger.debug(f"Mounting: {workspace_root_path_str}")
            logger.debug(f"\tto: {workspace_root_container_path_str}")
            devbox_volumes[workspace_root_path_str] = {
                "bind": workspace_root_container_path_str,
                "mode": "rw",
            }
        # Create a volume for Aws credentials
        if self.args.mount_aws_config:
            aws_config_path_str = str(self.args.aws_config_path)
            logger.debug(f"Mounting: {aws_config_path_str}")
            logger.debug(f"\tto: {self.args.aws_config_container_path}")
            devbox_volumes[aws_config_path_str] = {
                "bind": self.args.aws_config_container_path,
                "mode": "ro",
            }
            devbox_container_env[
                "AWS_CONFIG_FILE"
            ] = f"{self.args.aws_config_container_path}/config"
            devbox_container_env[
                "AWS_SHARED_CREDENTIALS_FILE"
            ] = f"{self.args.aws_config_container_path}/credentials"
        # Create a volume for Phidata dev mode
        # if dev_mode is enabled
        if self.args.dev_mode is not None:
            # mount the phidata directory if install_phidata_dev = True
            if (
                self.args.dev_mode.install_phidata_dev
                and self.args.dev_mode.phidata_dir_path is not None
            ):
                phidata_dir_absolute_path_str = str(self.args.dev_mode.phidata_dir_path)
                logger.debug(f"Mounting: {phidata_dir_absolute_path_str}")
                logger.debug(f"\tto: {self.args.dev_mode.phidata_dir_container_path}")
                devbox_volumes[phidata_dir_absolute_path_str] = {
                    "bind": self.args.dev_mode.phidata_dir_container_path,
                    "mode": "rw",
                }

        # Create the devbox container
        devbox_docker_container = DockerContainer(
            name=self.args.name,
            image=f"{self.args.image_name}:{self.args.image_tag}",
            command=self.args.command,
            auto_remove=True,
            detach=True,
            entrypoint=self.args.entrypoint,
            environment=devbox_container_env,
            remove=True,
            stdin_open=True,
            tty=True,
            network=docker_build_context.network,
            volumes=devbox_volumes,
            use_cache=self.args.use_cache,
            use_verbose_logs=self.args.use_verbose_logs,
        )
        # Initialize airflow on container
        self.init_airflow_on_docker_container(devbox_docker_container)
        # logger.debug(f"Devbox Container Env: {devbox_docker_container.environment}")

        devbox_docker_resource_group = DockerResourceGroup(
            name=self.args.name,
            enabled=self.args.enabled,
            network=DockerNetwork(name=docker_build_context.network),
            containers=[devbox_docker_container],
        )
        return devbox_docker_resource_group

    def init_docker_resource_groups(
        self, docker_build_context: DockerBuildContext
    ) -> None:
        devbox_docker_rg: Optional[DockerResourceGroup] = self.get_devbox_docker_rg(
            docker_build_context
        )
        if devbox_docker_rg is not None:
            if self.docker_resource_groups is None:
                self.docker_resource_groups = OrderedDict()
            self.docker_resource_groups[devbox_docker_rg.name] = devbox_docker_rg

    ######################################################
    ## K8s Resources
    ######################################################

    def init_airflow_on_k8s_container(
        self, container: CreateContainer, k8s_resource_group: CreateK8sResourceGroup
    ) -> None:
        """
        Initialize airflow on a k8s container
        """

        if not self.args.init_airflow:
            return

        # Update init_airflow_db arg if None
        if self.args.init_airflow_db is None:
            self.args.init_airflow_db = True

        airflow_cm = CreateConfigMap(
            cm_name=get_default_configmap_name("devbox-airflow"),
            app_name=self.args.name,
            data={
                "INIT_AIRFLOW": str(self.args.init_airflow),
                "INIT_AIRFLOW_DB": str(self.args.init_airflow_db),
                "WAIT_FOR_AIRFLOW_DB": str(self.args.wait_for_airflow_db),
                "AIRFLOW_DB_CONN_URL": self.args.airflow_db_conn_url,
                "AIRFLOW_DB_CONN_PORT": self.args.airflow_db_conn_port,
                "AIRFLOW_DB_USER": self.args.airflow_db_user,
                "AIRFLOW_DB_PASSWORD": self.args.airflow_db_password,
                "AIRFLOW_SCHEMA": self.args.airflow_schema,
                "INIT_AIRFLOW_SCHEDULER": str(self.args.init_airflow_scheduler),
                "INIT_AIRFLOW_WEBSERVER": str(self.args.init_airflow_webserver),
                "CREATE_AIRFLOW_TEST_USER": str(self.args.init_airflow_webserver),
                "AIRFLOW__CORE__LOAD_EXAMPLES": str(self.args.load_examples),
            },
        )

        # Add airflow_cm to container env and
        if container.envs_from_configmap is None:
            container.envs_from_configmap = []
        container.envs_from_configmap.append(airflow_cm.cm_name)
        if k8s_resource_group.config_maps is None:
            k8s_resource_group.config_maps = []
        k8s_resource_group.config_maps.append(airflow_cm)

        # Set the AIRFLOW__CORE__DAGS_FOLDER
        if self.args.mount_workspace and self.args.use_products_as_airflow_dags:
            projects_dir_container_path = Path(
                self.args.workspace_parent_container_path
            ).joinpath(self.args.workspace_dir_path.stem, self.args.projects_dir)
            airflow_cm.data["AIRFLOW__CORE__DAGS_FOLDER"] = projects_dir_container_path
        elif self.args.airflow_dags_path is not None:
            airflow_cm.data["AIRFLOW__CORE__DAGS_FOLDER"] = self.args.airflow_dags_path
        if self.args.db_connections is not None:
            for conn_id, conn_url in self.args.db_connections.items():
                try:
                    af_conn_id = str("AIRFLOW_CONN_{}".format(conn_id)).upper()
                    airflow_cm.data[af_conn_id] = conn_url
                except Exception as e:
                    logger.exception(e)
                    continue
        # Open the airflow webserver port if init_airflow_webserver = True
        if self.args.init_airflow_webserver:
            ws_port = CreatePort(
                name=self.args.airflow_webserver_port_name,
                container_port=self.args.airflow_webserver_container_port,
            )
            if container.ports is None:
                container.ports = []
            container.ports.append(ws_port)

    def get_devbox_k8s_rg(
        self, k8s_build_context: K8sBuildContext
    ) -> Optional[K8sResourceGroup]:
        logger.debug(f"Init Devbox K8sResourceGroup")

        # Set the Devbox Env
        # Devbox Config-Map is a dict of environment variables which are not secrets.
        # The keys of this configmap will become env variables in the devbox container
        workspace_dir_container_path = Path(
            self.args.workspace_parent_container_path
        ).joinpath(self.args.workspace_dir_path.stem)
        projects_dir_container_path = workspace_dir_container_path.joinpath(
            self.args.projects_dir
        )
        devbox_cm = CreateConfigMap(
            cm_name=get_default_configmap_name(self.args.name),
            app_name=self.args.name,
            data={
                # Env variables used by data workflows and data assets
                "PHI_DATA_DIR": self.args.data_dir_container_path,
                "PHI_PROJECTS_DIR": str(projects_dir_container_path),
                "PHI_WORKSPACE_DIR": str(workspace_dir_container_path),
                "PHI_WORKSPACE_PARENT": str(self.args.workspace_parent_container_path),
                # Configure python packages to install
                "INSTALL_REQUIREMENTS": str(self.args.install_requirements),
                "REQUIREMENTS_FILE_PATH": self.args.requirements_file_container_path,
                # Print these env variables when the container is loaded
                "PRINT_ENV_ON_LOAD": str(self.args.print_env_on_load),
                # Dev mode args
                "INSTALL_PHIDATA_DEV": str(self.args.dev_mode.install_phidata_dev)
                if self.args.dev_mode is not None
                else str(False),
                "PHIDATA_DIR_PATH": self.args.dev_mode.phidata_dir_container_path
                if self.args.dev_mode is not None
                else str(False),
            },
        )
        # Update the devbox env with user provided env
        if self.args.env is not None and isinstance(self.args.env, dict):
            devbox_cm.data.update(self.args.env)

        # Devbox Volumes
        devbox_volumes = []
        # Create a volume for the requirements file
        # if install_requirements = True
        if self.args.install_requirements:
            requirements_file_absolute_path = str(
                self.args.workspace_dir_path.joinpath(self.args.requirements_file_path)
            )
            logger.debug(f"Mounting: {requirements_file_absolute_path}")
            logger.debug(f"\tto: {self.args.requirements_file_container_path}")
            requirements_volume = CreateVolume(
                volume_name=self.args.requirements_volume_name,
                app_name=self.args.name,
                mount_path=self.args.requirements_file_container_path,
                volume_type=VolumeType.HOST_PATH,
                host_path=HostPathVolumeSource(path=requirements_file_absolute_path),
            )
            devbox_volumes.append(requirements_volume)
        # Create a volume for the data dir
        # if create_data_volume = True
        if self.args.create_data_volume:
            if (
                self.args.use_host_path_for_data_volume
                and self.args.data_dir_path is not None
            ):
                data_dir_absolute_path = str(
                    self.args.workspace_dir_path.joinpath(self.args.data_dir_path)
                )
                logger.debug(f"Mounting: {data_dir_absolute_path}")
                logger.debug(f"\tto: {self.args.data_dir_container_path}")
                data_volume = CreateVolume(
                    volume_name=self.args.data_dir_volume_name,
                    app_name=self.args.name,
                    mount_path=self.args.data_dir_container_path,
                    volume_type=VolumeType.HOST_PATH,
                    host_path=HostPathVolumeSource(
                        path=data_dir_absolute_path,
                    ),
                )
                devbox_volumes.append(data_volume)
            elif self.args.use_empty_dir_for_data_volume:
                logger.debug(f"Mounting EmptyDir")
                logger.debug(f"\tto: {self.args.data_dir_container_path}")
                data_volume = CreateVolume(
                    volume_name=self.args.data_dir_volume_name,
                    app_name=self.args.name,
                    mount_path=self.args.data_dir_container_path,
                    volume_type=VolumeType.EMPTY_DIR,
                    empty_dir=EmptyDirVolumeSource(),
                )
                devbox_volumes.append(data_volume)
        # Create a volume for the projects dir
        # if mount_workspace = True
        if self.args.mount_workspace:
            if self.args.use_host_path_for_workspace:
                ws_dir_container_path_str = str(
                    Path(self.args.workspace_parent_container_path).joinpath(
                        self.args.workspace_dir_path.stem
                    )
                )
                ws_dir_path_str = str(self.args.workspace_dir_path)
                logger.debug(f"Mounting: {ws_dir_path_str}")
                logger.debug(f"\tto: {ws_dir_container_path_str}")
                workspace_volume = CreateVolume(
                    volume_name=self.args.workspace_volume_name,
                    app_name=self.args.name,
                    mount_path=ws_dir_container_path_str,
                    volume_type=VolumeType.HOST_PATH,
                    host_path=HostPathVolumeSource(
                        path=ws_dir_path_str,
                    ),
                )
                devbox_volumes.append(workspace_volume)
        # Airflow conf file
        if self.args.mount_airflow_dir:
            airflow_dir_file_absolute_path = str(
                self.args.workspace_dir_path.joinpath(self.args.airflow_dir_path)
            )
            logger.debug(f"Mounting: {airflow_dir_file_absolute_path}")
            logger.debug(f"\tto: {self.args.airflow_dir_container_path}")
            airflow_dir_volume = CreateVolume(
                volume_name=self.args.airflow_volume_name,
                app_name=self.args.name,
                mount_path=self.args.airflow_dir_container_path,
                volume_type=VolumeType.HOST_PATH,
                host_path=HostPathVolumeSource(path=airflow_dir_file_absolute_path),
            )
            devbox_volumes.append(airflow_dir_volume)
        # Create a volume for Phidata dev mode
        # if dev_mode is enabled
        if self.args.dev_mode is not None:
            # mount the phidata directory if install_phidata_dev = True
            if (
                self.args.dev_mode.install_phidata_dev
                and self.args.dev_mode.phidata_dir_path is not None
            ):
                phidata_dir_absolute_path = str(self.args.dev_mode.phidata_dir_path)
                logger.debug(f"Mounting: {phidata_dir_absolute_path}")
                logger.debug(f"\tto: {self.args.dev_mode.phidata_dir_container_path}")
                phidata_volume = CreateVolume(
                    volume_name=self.args.dev_mode.phidata_volume_name,
                    app_name=self.args.name,
                    mount_path=self.args.dev_mode.phidata_dir_container_path,
                    volume_type=VolumeType.HOST_PATH,
                    host_path=HostPathVolumeSource(
                        path=phidata_dir_absolute_path,
                    ),
                )
                devbox_volumes.append(phidata_volume)

        # Create the devbox container
        devbox_k8s_container = CreateContainer(
            container_name=f"{self.args.name}-container",
            app_name=self.args.name,
            image_name=self.args.image_name,
            image_tag=self.args.image_tag,
            # Equivalent to docker images CMD
            args=self.args.command,
            # Equivalent to docker images ENTRYPOINT
            command=self.args.entrypoint,
            envs_from_configmap=[devbox_cm.cm_name],
            volumes=devbox_volumes,
        )
        devbox_deployment = CreateDeployment(
            deploy_name=f"{self.args.name}-deploy",
            pod_name=f"{self.args.name}-pod",
            app_name=self.args.name,
            namespace=k8s_build_context.namespace,
            service_account_name=k8s_build_context.service_account_name,
            containers=[devbox_k8s_container],
            volumes=devbox_volumes,
            labels=k8s_build_context.labels,
        )
        devbox_k8s_resource_group = CreateK8sResourceGroup(
            name=self.args.name,
            enabled=self.args.enabled,
            config_maps=[devbox_cm],
            deployments=[devbox_deployment],
        )
        # Initialize airflow on container
        self.init_airflow_on_k8s_container(
            devbox_k8s_container, devbox_k8s_resource_group
        )
        return devbox_k8s_resource_group.create()

    def init_k8s_resource_groups(self, k8s_build_context: K8sBuildContext) -> None:
        devbox_k8s_rg: Optional[K8sResourceGroup] = self.get_devbox_k8s_rg(
            k8s_build_context
        )
        # logger.debug("devbox_k8s_rg: {}".format(devbox_k8s_rg))
        if devbox_k8s_rg is not None:
            if self.k8s_resource_groups is None:
                self.k8s_resource_groups = OrderedDict()
            self.k8s_resource_groups[devbox_k8s_rg.name] = devbox_k8s_rg
